@@ -1,6 +1,16 @@
 #ifndef __OPTIONS_H__
 #define __OPTIONS_H__
 
+#include "local_path.h"
+
+#include <option_change_event_handler.h>
+
+#include <libfilezilla/mutex.hpp>
+
+#include <wx/timer.h>
+
+#include "xmlfunctions.h"
+
 enum interfaceOptions
 {
 	OPTION_NUMTRANSFERS = OPTIONS_ENGINE_NUM,
@@ -18,7 +28,6 @@ enum interfaceOptions
 	OPTION_UPDATECHECK_LASTDATE,
 	OPTION_UPDATECHECK_NEWVERSION,
 	OPTION_UPDATECHECK_CHECKBETA,
-	OPTION_UPDATECHECK_DOWNLOADDIR,
 	OPTION_DEBUG_MENU,
 	OPTION_FILEEXISTS_DOWNLOAD,
 	OPTION_FILEEXISTS_UPLOAD,
@@ -31,6 +40,7 @@ enum interfaceOptions
 	OPTION_FILEPANE_SWAP,
 	OPTION_LASTLOCALDIR,
 	OPTION_FILELIST_DIRSORT,
+	OPTION_FILELIST_NAMESORT,
 	OPTION_QUEUE_SUCCESSFUL_AUTOCLEAR,
 	OPTION_QUEUE_COLUMN_WIDTHS,
 	OPTION_LOCALFILELIST_COLUMN_WIDTHS,
@@ -80,9 +90,13 @@ enum interfaceOptions
 	OPTION_EDITSTATUSDIALOG_SIZE,
 	OPTION_SPEED_DISPLAY,
 	OPTION_TOOLBAR_HIDDEN,
+	OPTION_STRIP_VMS_REVISION,
+	OPTION_INTERFACE_SITEMANAGER_ON_STARTUP,
+	OPTION_PROMPTPASSWORDSAVE,
+	OPTION_PERSISTENT_CHOICES,
 
 	// Default/internal options
-	OPTION_DEFAULT_SETTINGSDIR,
+	OPTION_DEFAULT_SETTINGSDIR, // guaranteed to be (back)slash-terminated
 	OPTION_DEFAULT_KIOSKMODE,
 	OPTION_DEFAULT_DISABLEUPDATECHECK,
 
@@ -92,13 +106,17 @@ enum interfaceOptions
 
 struct t_OptionsCache
 {
+	bool operator==(wxString const& v) const { return strValue == v; }
+	bool operator==(int v) const { return numValue == v; }
+	t_OptionsCache& operator=(wxString const& v);
+	t_OptionsCache& operator=(int v);
+
 	bool from_default;
 	int numValue;
 	wxString strValue;
 };
 
 class CXmlFile;
-class TiXmlElement;
 class COptions : public wxEvtHandler, public COptionsBase
 {
 public:
@@ -106,10 +124,10 @@ public:
 	virtual wxString GetOption(unsigned int nID);
 
 	virtual bool SetOption(unsigned int nID, int value);
-	virtual bool SetOption(unsigned int nID, wxString value);
+	virtual bool SetOption(unsigned int nID, wxString const& value);
 
 	bool OptionFromFzDefaultsXml(unsigned int nID);
-	
+
 	void SetLastServer(const CServer& server);
 	bool GetLastServer(CServer& server);
 
@@ -117,38 +135,43 @@ public:
 	static void Init();
 	static void Destroy();
 
-	void Import(TiXmlElement* pElement);
+	void Import(pugi::xml_node element);
 
 	void SaveIfNeeded();
 
+	static CLocalPath GetUnadjustedSettingsDir();
+
 protected:
 	COptions();
 	virtual ~COptions();
 
 	int Validate(unsigned int nID, int value);
-	wxString Validate(unsigned int nID, wxString value);
+	wxString Validate(unsigned int nID, wxString const& value);
 
-	void SetXmlValue(unsigned int nID, wxString value);
+	template<typename T> void ContinueSetOption(unsigned int nID, T const& value);
+	void SetXmlValue(unsigned int nID, int value);
+	void SetXmlValue(unsigned int nID, wxString const& value);
 
 	// path is element path below document root, separated by slashes
 	void SetServer(wxString path, const CServer& server);
 	bool GetServer(wxString path, CServer& server);
 
-	void CreateSettingsXmlElement();
+	pugi::xml_node CreateSettingsXmlElement();
 
-	void GetNameOptionMap(std::map<std::string, int>& nameOptionMap) const; //Give me C++1x
-	void LoadOptions(const std::map<std::string, int>& nameOptionMap, TiXmlElement* settings = 0);
-	void LoadGlobalDefaultOptions(const std::map<std::string, int>& nameOptionMap);
-	void LoadOptionFromElement(TiXmlElement* pOption, const std::map<std::string, int>& nameOptionMap, bool allowDefault);
-	void InitSettingsDir();
+	std::map<std::string, unsigned int> GetNameOptionMap() const;
+	void LoadOptions(std::map<std::string, unsigned int> const& nameOptionMap, pugi::xml_node settings = pugi::xml_node());
+	void LoadGlobalDefaultOptions(std::map<std::string, unsigned int> const& nameOptionMap);
+	void LoadOptionFromElement(pugi::xml_node option, std::map<std::string, unsigned int> const& nameOptionMap, bool allowDefault);
+	CLocalPath InitSettingsDir();
 	void SetDefaultValues();
 
 	void Save();
 
+	void NotifyChangedOptions();
+
 	CXmlFile* m_pXmlFile;
 
 	t_OptionsCache m_optionsCache[OPTIONS_NUM];
-	bool m_acquired;
 
 	CServer* m_pLastServer;
 
@@ -156,8 +179,12 @@ protected:
 
 	wxTimer m_save_timer;
 
-	DECLARE_EVENT_TABLE();
+	DECLARE_EVENT_TABLE()
 	void OnTimer(wxTimerEvent& event);
+
+	fz::mutex m_sync_;
+
+	changed_options_t changedOptions_;
 };
 
 #endif //__OPTIONS_H__
