@@ -21,7 +21,6 @@
 #include <wx/utils.h>
 #include <wx/progdlg.h>
 #include <wx/sound.h>
-#include "local_filesys.h"
 #include "statusbar.h"
 #include "recursive_operation.h"
 #include "auto_ascii_files.h"
@@ -29,8 +28,12 @@
 #include "drop_target_ex.h"
 #if WITH_LIBDBUS
 #include "../dbus/desktop_notification.h"
+#elif defined(__WXGTK__) || defined(__WXMSW__)
+#include <wx/notifmsg.h>
 #endif
 
+#include <libfilezilla/local_filesys.hpp>
+
 #ifdef __WXMSW__
 #include <powrprof.h>
 #endif
@@ -209,7 +212,7 @@ EVT_COMMAND(wxID_ANY, fzEVT_GRANTEXCLUSIVEENGINEACCESS, CQueueView::OnExclusiveE
 EVT_SIZE(CQueueView::OnSize)
 END_EVENT_TABLE()
 
-class CFolderProcessingThread final : public wxThread
+class CFolderProcessingThread final : public fz::thread
 {
 	struct t_internalDirPair
 	{
@@ -218,15 +221,10 @@ class CFolderProcessingThread final : public wxThread
 	};
 public:
 	CFolderProcessingThread(CQueueView* pOwner, CFolderScanItem* pFolderItem)
-		: wxThread(wxTHREAD_JOINABLE) {
+	{
 		m_pOwner = pOwner;
 		m_pFolderItem = pFolderItem;
 
-		m_didSendEvent = false;
-		m_threadWaiting = false;
-		m_throttleWait = false;
-		m_processing_entries = false;
-
 		t_internalDirPair* pair = new t_internalDirPair;
 		pair->localPath = pFolderItem->GetLocalPath();
 		pair->remotePath = pFolderItem->GetRemotePath();
@@ -235,6 +233,11 @@ public:
 
 	virtual ~CFolderProcessingThread()
 	{
+		{
+			fz::scoped_lock locker(m_sync);
+			m_quit = true;
+		}
+		join();
 		for (auto iter = m_entryList.begin(); iter != m_entryList.end(); ++iter)
 			delete *iter;
 		for (auto iter = m_dirsToCheck.begin(); iter != m_dirsToCheck.end(); ++iter)
@@ -244,7 +247,7 @@ public:
 	void GetFiles(std::list<CFolderProcessingEntry*> &entryList)
 	{
 		wxASSERT(entryList.empty());
-		scoped_lock locker(m_sync);
+		fz::scoped_lock locker(m_sync);
 		entryList.swap(m_entryList);
 
 		m_didSendEvent = false;
@@ -266,7 +269,7 @@ public:
 
 	void ProcessDirectory(const CLocalPath& localPath, CServerPath const& remotePath, const wxString& name)
 	{
-		scoped_lock locker(m_sync);
+		fz::scoped_lock locker(m_sync);
 
 		t_internalDirPair* pair = new t_internalDirPair;
 
@@ -288,7 +291,7 @@ public:
 
 	void CheckFinished()
 	{
-		scoped_lock locker(m_sync);
+		fz::scoped_lock locker(m_sync);
 		wxASSERT(m_processing_entries);
 
 		m_processing_entries = false;
@@ -308,7 +311,7 @@ protected:
 
 	void AddEntry(CFolderProcessingEntry* entry)
 	{
-		scoped_lock l(m_sync);
+		fz::scoped_lock l(m_sync);
 		m_entryList.push_back(entry);
 
 		// Wait if there are more than 100 items to queue,
@@ -341,7 +344,7 @@ protected:
 		}
 	}
 
-	ExitCode Entry()
+	void entry()
 	{
 #ifdef __WXDEBUG__
 		wxMutexGuiEnter();
@@ -351,10 +354,13 @@ protected:
 
 		wxASSERT(!m_pFolderItem->Download());
 
-		CLocalFileSystem localFileSystem;
+		fz::local_filesys localFileSystem;
 
-		while (!TestDestroy() && !m_pFolderItem->m_remove) {
-			scoped_lock l(m_sync);
+		while (!m_pFolderItem->m_remove) {
+			fz::scoped_lock l(m_sync);
+			if (m_quit) {
+				break;
+			}
 			if (m_dirsToCheck.empty()) {
 				if (!m_didSendEvent && !m_entryList.empty()) {
 					m_didSendEvent = true;
@@ -379,7 +385,7 @@ protected:
 
 			l.unlock();
 
-			if (!localFileSystem.BeginFindFiles(pair->localPath.GetPath(), false)) {
+			if (!localFileSystem.begin_find_files(fz::to_native(pair->localPath.GetPath()), false)) {
 				delete pair;
 				continue;
 			}
@@ -391,10 +397,10 @@ protected:
 
 			t_newEntry* entry = new t_newEntry;
 
-			wxString name;
+			fz::native_string name;
 			bool is_link;
 			bool is_dir;
-			while (localFileSystem.GetNextFile(name, is_link, is_dir, &entry->size, &entry->time, &entry->attributes)) {
+			while (localFileSystem.get_next_file(name, is_link, is_dir, &entry->size, &entry->time, &entry->attributes)) {
 				if (is_link)
 					continue;
 
@@ -411,7 +417,6 @@ protected:
 		}
 
 		m_pOwner->QueueEvent(new wxCommandEvent(fzEVT_FOLDERTHREAD_COMPLETE, wxID_ANY));
-		return 0;
 	}
 
 	std::list<t_internalDirPair*> m_dirsToCheck;
@@ -422,12 +427,13 @@ protected:
 	CQueueView* m_pOwner;
 	CFolderScanItem* m_pFolderItem;
 
-	mutex m_sync;
-	condition m_condition;
-	bool m_threadWaiting;
-	bool m_throttleWait;
-	bool m_didSendEvent;
-	bool m_processing_entries;
+	fz::mutex m_sync;
+	fz::condition m_condition;
+	bool m_threadWaiting{};
+	bool m_throttleWait{};
+	bool m_didSendEvent{};
+	bool m_processing_entries{};
+	bool m_quit{};
 };
 
 CQueueView::CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsyncRequestQueue *pAsyncRequestQueue)
@@ -445,8 +451,7 @@ CQueueView::CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsync
 	m_quit = 0;
 	m_waitStatusLineUpdate = false;
 	m_lastTopItem = -1;
-	m_pFolderProcessingThread = 0;
-
+	
 	m_totalQueueSize = 0;
 	m_filesWithUnknownSize = 0;
 
@@ -470,26 +475,16 @@ CQueueView::CQueueView(CQueue* parent, int index, CMainFrame* pMainFrame, CAsync
 #endif
 
 	m_resize_timer.SetOwner(this);
-
-#if WITH_LIBDBUS
-	m_desktop_notification = 0;
-#endif
 }
 
 CQueueView::~CQueueView()
 {
-	if (m_pFolderProcessingThread) {
-		m_pFolderProcessingThread->Delete(0, wxTHREAD_WAIT_BLOCK);
-		delete m_pFolderProcessingThread;
-	}
+	delete m_pFolderProcessingThread;
+	m_pFolderProcessingThread = 0;
 
 	DeleteEngines();
 
 	m_resize_timer.Stop();
-
-#if WITH_LIBDBUS
-	delete m_desktop_notification;
-#endif
 }
 
 bool CQueueView::QueueFile(const bool queueOnly, const bool download,
@@ -871,12 +866,9 @@ bool CQueueView::TryStartNextTransfer()
 	const CServer oldServer = pEngineData->lastServer;
 	pEngineData->lastServer = bestMatch.serverItem->GetServer();
 
-	if (pEngineData->state != t_EngineData::waitprimary)
-	{
-		if (!pEngineData->pEngine->IsConnected())
-		{
-			if (pEngineData->lastServer.GetLogonType() == ASK)
-			{
+	if (pEngineData->state != t_EngineData::waitprimary) {
+		if (!pEngineData->pEngine->IsConnected()) {
+			if (pEngineData->lastServer.GetLogonType() == ASK) {
 				if (CLoginManager::Get().GetPassword(pEngineData->lastServer, true))
 					pEngineData->state = t_EngineData::connect;
 				else
@@ -893,8 +885,7 @@ bool CQueueView::TryStartNextTransfer()
 			pEngineData->state = t_EngineData::mkdir;
 	}
 
-	if (bestMatch.fileItem->GetType() == QueueItemType::File)
-	{
+	if (bestMatch.fileItem->GetType() == QueueItemType::File) {
 		// Create status line
 
 		m_itemCount++;
@@ -1103,14 +1094,12 @@ void CQueueView::ProcessReply(t_EngineData* pEngineData, COperationNotification
 		return;
 	}
 
-	if (pEngineData->state == t_EngineData::connect && pEngineData->lastServer.GetLogonType() == ASK)
-	{
+	if (pEngineData->state == t_EngineData::connect && pEngineData->lastServer.GetLogonType() == ASK) {
 		if (!CLoginManager::Get().GetPassword(pEngineData->lastServer, true))
 			pEngineData->state = t_EngineData::askpassword;
 	}
 
-	if (!m_activeMode)
-	{
+	if (!m_activeMode) {
 		enum ResetReason reason;
 		if (pEngineData->pItem && pEngineData->pItem->pending_remove())
 			reason = remove;
@@ -1266,7 +1255,7 @@ void CQueueView::ResetEngine(t_EngineData& data, const enum ResetReason reason)
 	UpdateStatusLinePositions();
 }
 
-bool CQueueView::RemoveItem(CQueueItem* item, bool destroy, bool updateItemCount /*=true*/, bool updateSelections /*=true*/)
+bool CQueueView::RemoveItem(CQueueItem* item, bool destroy, bool updateItemCount, bool updateSelections, bool forward)
 {
 	// RemoveItem assumes that the item has already been removed from all engines
 
@@ -1289,7 +1278,7 @@ bool CQueueView::RemoveItem(CQueueItem* item, bool destroy, bool updateItemCount
 		}
 	}
 
-	bool didRemoveParent = CQueueViewBase::RemoveItem(item, destroy, updateItemCount, updateSelections);
+	bool didRemoveParent = CQueueViewBase::RemoveItem(item, destroy, updateItemCount, updateSelections, forward);
 
 	UpdateStatusLinePositions();
 
@@ -1620,12 +1609,10 @@ void CQueueView::UpdateStatusLinePositions()
 	m_lastTopItem = GetTopItem();
 	int bottomItem = m_lastTopItem + GetCountPerPage();
 
-	for (auto iter = m_statusLineList.begin(); iter != m_statusLineList.end(); ++iter)
-	{
+	for (auto iter = m_statusLineList.begin(); iter != m_statusLineList.end(); ++iter) {
 		CStatusLineCtrl *pCtrl = *iter;
 		int index = GetItemIndex(pCtrl->GetItem()) + 1;
-		if (index < m_lastTopItem || index > bottomItem)
-		{
+		if (index < m_lastTopItem || index > bottomItem) {
 			pCtrl->Show(false);
 			continue;
 		}
@@ -1701,8 +1688,7 @@ bool CQueueView::ProcessFolderItems(int type /*=-1*/)
 
 void CQueueView::ProcessUploadFolderItems()
 {
-	if (m_queuedFolders[1].empty())
-	{
+	if (m_queuedFolders[1].empty()) {
 		if (m_quit)
 			m_pMainFrame->Close();
 
@@ -1721,8 +1707,7 @@ void CQueueView::ProcessUploadFolderItems()
 	RefreshItem(pItem);
 	pItem->m_active = true;
 	m_pFolderProcessingThread = new CFolderProcessingThread(this, pItem);
-	m_pFolderProcessingThread->Create();
-	m_pFolderProcessingThread->Run();
+	m_pFolderProcessingThread->run();
 
 	RefreshListOnly(false);
 }
@@ -1736,8 +1721,7 @@ void CQueueView::OnFolderThreadComplete(wxCommandEvent&)
 
 	wxASSERT(!m_queuedFolders[1].empty());
 	CFolderScanItem* pItem = m_queuedFolders[1].front();
-	if (pItem->m_dir_is_empty)
-	{
+	if (pItem->m_dir_is_empty) {
 		CServerItem* pServerItem = (CServerItem*)pItem->GetTopLevelItem();
 		CFileItem* fileItem = new CFolderItem(pServerItem, pItem->queued(), pItem->m_current_remote_path, _T(""));
 		InsertItem(pServerItem, fileItem);
@@ -1747,7 +1731,6 @@ void CQueueView::OnFolderThreadComplete(wxCommandEvent&)
 
 	RemoveItem(pItem, true);
 
-	m_pFolderProcessingThread->Wait(wxTHREAD_WAIT_BLOCK);
 	delete m_pFolderProcessingThread;
 	m_pFolderProcessingThread = 0;
 
@@ -2220,42 +2203,53 @@ void CQueueView::OnRemoveSelected(wxCommandEvent&)
 		return;
 #endif
 
-	std::list<CQueueItem*> selectedItems;
+	std::list<std::pair<long, CQueueItem*>> selectedItems;
 	long item = -1;
-	for (;;)
-	{
+	long skipTo = -1;
+	for (;;) {
 		item = GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
 		if (item == -1)
 			break;
-
-		selectedItems.push_front(GetQueueItem(item));
 		SetItemState(item, 0, wxLIST_STATE_SELECTED);
+
+		if (item <= skipTo) {
+			continue;
+		}
+
+		CQueueItem* pItem = GetQueueItem(item);
+		if (!pItem) {
+			continue;
+		}
+
+		selectedItems.push_front(std::make_pair(item, pItem));
+
+		if (pItem->GetType() == QueueItemType::Server) {
+			// Server selected. Don't process individual files, continue with the next server
+			skipTo = item + pItem->GetChildrenCount(true);
+		}
 	}
 
 	m_waitStatusLineUpdate = true;
 
-	while (!selectedItems.empty())
-	{
-		CQueueItem* pItem = selectedItems.front();
+	while (!selectedItems.empty()) {
+		auto selectedItem = selectedItems.front();
+		CQueueItem* pItem = selectedItem.second;
 		selectedItems.pop_front();
 
 		if (pItem->GetType() == QueueItemType::Status)
 			continue;
-		else if (pItem->GetType() == QueueItemType::FolderScan)
-		{
+		else if (pItem->GetType() == QueueItemType::FolderScan) {
 			CFolderScanItem* pFolder = (CFolderScanItem*)pItem;
-			if (pFolder->m_active)
-			{
+			if (pFolder->m_active) {
 				pFolder->m_remove = true;
 				continue;
 			}
 			else
 				RemoveQueuedFolderItem(pFolder);
 		}
-		else if (pItem->GetType() == QueueItemType::Server)
-		{
+		else if (pItem->GetType() == QueueItemType::Server) {
 			CServerItem* pServer = (CServerItem*)pItem;
-			StopItem(pServer);
+			StopItem(pServer, false);
 
 			// Server items get deleted automatically if all children are gone
 			continue;
@@ -2264,8 +2258,7 @@ void CQueueView::OnRemoveSelected(wxCommandEvent&)
 				 pItem->GetType() == QueueItemType::Folder)
 		{
 			CFileItem* pFile = (CFileItem*)pItem;
-			if (pFile->IsActive())
-			{
+			if (pFile->IsActive()) {
 				pFile->set_pending_remove(true);
 				StopItem(pFile);
 				continue;
@@ -2273,14 +2266,19 @@ void CQueueView::OnRemoveSelected(wxCommandEvent&)
 		}
 
 		CQueueItem* pTopLevelItem = pItem->GetTopLevelItem();
-		if (!pTopLevelItem->GetChild(1))
-		{
+		if (!pTopLevelItem->GetChild(1)) {
 			// Parent will get deleted
 			// If next selected item is parent, remove it from list
-			if (!selectedItems.empty() && selectedItems.front() == pTopLevelItem)
+			if (!selectedItems.empty() && selectedItems.front().second == pTopLevelItem)
 				selectedItems.pop_front();
 		}
-		RemoveItem(pItem, true, false, false);
+
+		int topItemIndex = GetItemIndex(pTopLevelItem);
+
+		// Finding the child position is O(n) in the worst case, making deletion quadratic. However we already know item's displayed position, use it as guide.
+		// Remark: I suppose this could be further improved by using the displayed position directly, but probably isn't worth the effort.
+		bool forward = selectedItem.first < (topItemIndex + static_cast<int>(pTopLevelItem->GetChildrenCount(false)) / 2);
+		RemoveItem(pItem, true, false, false, forward);
 	}
 	DisplayNumberQueuedFiles();
 	DisplayQueueSize();
@@ -2318,20 +2316,16 @@ bool CQueueView::StopItem(CFileItem* item)
 	}
 }
 
-bool CQueueView::StopItem(CServerItem* pServerItem)
+bool CQueueView::StopItem(CServerItem* pServerItem, bool updateSelections)
 {
-	std::list<CQueueItem*> items;
-	for (unsigned int i = 0; i < pServerItem->GetChildrenCount(false); i++)
-		items.push_back(pServerItem->GetChild(i, false));
+	std::vector<CQueueItem*> const items = pServerItem->GetChildren();
+	int const removedAtFront = pServerItem->GetRemovedAtFront();
 
-	for (std::list<CQueueItem*>::reverse_iterator iter = items.rbegin(); iter != items.rend(); ++iter)
-	{
-		CQueueItem* pItem = *iter;
-		if (pItem->GetType() == QueueItemType::FolderScan)
-		{
+	for (int i = static_cast<int>(items.size()) - 1; i >= removedAtFront; --i) {
+		CQueueItem* pItem = items[i];
+		if (pItem->GetType() == QueueItemType::FolderScan) {
 			CFolderScanItem* pFolder = (CFolderScanItem*)pItem;
-			if (pFolder->m_active)
-			{
+			if (pFolder->m_active) {
 				pFolder->m_remove = true;
 				continue;
 			}
@@ -2340,8 +2334,7 @@ bool CQueueView::StopItem(CServerItem* pServerItem)
 				 pItem->GetType() == QueueItemType::Folder)
 		{
 			CFileItem* pFile = (CFileItem*)pItem;
-			if (pFile->IsActive())
-			{
+			if (pFile->IsActive()) {
 				pFile->set_pending_remove(true);
 				StopItem(pFile);
 				continue;
@@ -2352,8 +2345,7 @@ bool CQueueView::StopItem(CServerItem* pServerItem)
 			wxASSERT(false);
 		}
 
-		if (RemoveItem(pItem, true, false))
-		{
+		if (RemoveItem(pItem, true, false, updateSelections, false)) {
 			DisplayNumberQueuedFiles();
 			SaveSetItemCount(m_itemCount);
 			return true;
@@ -2638,25 +2630,21 @@ void CQueueView::TryRefreshListings()
 
 void CQueueView::OnAskPassword(wxCommandEvent&)
 {
-	while (!m_waitingForPassword.empty())
-	{
+	while (!m_waitingForPassword.empty()) {
 		const CFileZillaEngine* const pEngine = m_waitingForPassword.front();
 
 		t_EngineData* pEngineData = GetEngineData(pEngine);
-		if (!pEngineData)
-		{
+		if (!pEngineData) {
 			m_waitingForPassword.pop_front();
 			continue;
 		}
 
-		if (pEngineData->state != t_EngineData::askpassword)
-		{
+		if (pEngineData->state != t_EngineData::askpassword) {
 			m_waitingForPassword.pop_front();
 			continue;
 		}
 
-		if (CLoginManager::Get().GetPassword(pEngineData->lastServer, false))
-		{
+		if (CLoginManager::Get().GetPassword(pEngineData->lastServer, false)) {
 			pEngineData->state = t_EngineData::connect;
 			SendNextCommand(*pEngineData);
 		}
@@ -2924,8 +2912,7 @@ void CQueueView::ActionAfter(bool warned /*=false*/)
 	// Need to check all contexts whether there's a recursive
 	// download operation still in progress
 	const std::vector<CState*> *pStates = CContextManager::Get()->GetAllStates();
-	for (unsigned int i = 0; i < pStates->size(); i++)
-	{
+	for (unsigned int i = 0; i < pStates->size(); ++i) {
 		CState *pState = (*pStates)[i];
 		CRecursiveOperation *pRecursiveOperationHandler;
 		if (!pState || !(pRecursiveOperationHandler = pState->GetRecursiveOperationHandler()))
@@ -2938,21 +2925,28 @@ void CQueueView::ActionAfter(bool warned /*=false*/)
 		}
 	}
 
-#if WITH_LIBDBUS
-	if (!m_pMainFrame->IsActive())
-	{
-		if (!m_desktop_notification)
-			m_desktop_notification = new CDesktopNotification;
-		int failed_count = m_pQueue->GetQueueView_Failed()->GetFileCount();
-		if (failed_count != 0)
-		{
+	if (!m_pMainFrame->IsActive()) {
+		wxString const title = _("Transfers finished");
+		wxString msg;
+		int const failed_count = m_pQueue->GetQueueView_Failed()->GetFileCount();
+		if (failed_count != 0) {
 			wxString fmt = wxPLURAL("All transfers have finished. %d file could not be transferred.", "All transfers have finished. %d files could not be transferred.", failed_count);
-			m_desktop_notification->Notify(_("Transfers finished"), wxString::Format(fmt, failed_count), _T("transfer.error"));
+			msg = wxString::Format(fmt, failed_count);
 		}
 		else
-			m_desktop_notification->Notify(_("Transfers finished"), _("All files have been successfully transferred"), _T("transfer.complete"));
-	}
+			msg = _("All files have been successfully transferred");
+
+#if WITH_LIBDBUS
+		if (!m_desktop_notification)
+			m_desktop_notification = std::make_unique<CDesktopNotification>();
+		m_desktop_notification->Notify(title, msg, (failed_count > 0) ? _T("transfer.error") : _T("transfer.complete"));
+#elif defined(__WXGTK__) || defined(__WXMSW__)
+		m_desktop_notification = std::make_unique<wxNotificationMessage>();
+		m_desktop_notification->SetTitle(title);
+		m_desktop_notification->SetMessage(msg);
+		m_desktop_notification->Show(5);
 #endif
+}
 
 	switch (m_actionAfterState)
 	{
