@@ -4,6 +4,8 @@
 #include <wx/aui/auibook.h>
 #include "state.h"
 
+#include <memory>
+
 class wxAuiNotebookEx;
 class CLocalListView;
 class CLocalTreeView;
@@ -15,7 +17,7 @@ class CViewHeader;
 class CSplitterWindowEx;
 class CState;
 
-class CContextControl : public wxSplitterWindow, public CStateEventHandler
+class CContextControl final : public wxSplitterWindow, public CStateEventHandler
 {
 public:
 	struct _context_controls
@@ -31,7 +33,7 @@ public:
 		CRemoteListView* pRemoteListView;
 		CViewHeader* pLocalViewHeader;
 		CViewHeader* pRemoteViewHeader;
-	
+
 		CSplitterWindowEx* pViewSplitter; // Contains local and remote splitters
 		CSplitterWindowEx* pLocalSplitter;
 		CSplitterWindowEx* pRemoteSplitter;
@@ -45,10 +47,10 @@ public:
 			wxString path;
 			std::list<wxString> bookmarks;
 		};
-		CSharedPointer<struct _site_bookmarks> site_bookmarks;
+		std::shared_ptr<_site_bookmarks> site_bookmarks;
 	};
 
-	CContextControl(CMainFrame* pMainFrame);
+	CContextControl(CMainFrame& mainFrame);
 	virtual ~CContextControl();
 
 	void Create(wxWindow* parent);
@@ -58,7 +60,7 @@ public:
 
 	struct _context_controls* GetCurrentControls();
 	struct _context_controls* GetControlsFromState(CState* pState);
-	
+
 	int GetCurrentTab() const;
 	int GetTabCount() const;
 	struct _context_controls* GetControlsFromTabIndex(int i);
@@ -67,29 +69,29 @@ public:
 	void AdvanceTab(bool forward);
 
 protected:
-	
+
 	void CreateContextControls(CState* pState);
 
 	std::vector<struct _context_controls> m_context_controls;
-	int m_current_context_controls;
+	int m_current_context_controls{-1};
 
-	wxAuiNotebookEx* m_tabs;
-	int m_right_clicked_tab;
-	CMainFrame* m_pMainFrame;
+	wxAuiNotebookEx* m_tabs{};
+	int m_right_clicked_tab{-1};
+	CMainFrame& m_mainFrame;
 
 protected:
-	DECLARE_EVENT_TABLE();
-	void OnTabRefresh(wxCommandEvent& event);
+	DECLARE_EVENT_TABLE()
+	void OnTabRefresh(wxCommandEvent&);
 	void OnTabChanged(wxAuiNotebookEvent& event);
 	void OnTabClosing(wxAuiNotebookEvent& event);
 	void OnTabClosing_Deferred(wxCommandEvent& event);
-	void OnTabBgDoubleclick(wxAuiNotebookEvent& event);
+	void OnTabBgDoubleclick(wxAuiNotebookEvent&);
 	void OnTabRightclick(wxAuiNotebookEvent& event);
 	void OnTabContextClose(wxCommandEvent& event);
 	void OnTabContextCloseOthers(wxCommandEvent& event);
-	void OnTabContextNew(wxCommandEvent& event);
+	void OnTabContextNew(wxCommandEvent&);
 
-	virtual void OnStateChange(CState* pState, enum t_statechange_notifications notification, const wxString& data, const void* data2);
+	virtual void OnStateChange(CState* pState, enum t_statechange_notifications notification, const wxString&, const void*);
 };
 
 #endif //__CONTEX_CONTROL_H__
