@@ -20,6 +20,34 @@ public:
 	CServerPath path;
 };
 
+template<>
+inline int DoCmpName(CSearchFileData const& data1, CSearchFileData const& data2, CFileListCtrlSortBase::NameSortMode const nameSortMode)
+{
+	int res;
+	switch (nameSortMode)
+	{
+	case CFileListCtrlSortBase::namesort_casesensitive:
+		res = CFileListCtrlSortBase::CmpCase(data1.name, data2.name);
+		break;
+	default:
+	case CFileListCtrlSortBase::namesort_caseinsensitive:
+		res = CFileListCtrlSortBase::CmpNoCase(data1.name, data2.name);
+		break;
+	case CFileListCtrlSortBase::namesort_natural:
+		res = CFileListCtrlSortBase::CmpNatural(data1.name, data2.name);
+		break;
+	}
+
+	if (!res) {
+		if (data1.path < data2.path)
+			res = -1;
+		else if (data2.path < data1.path)
+			res = 1;
+	}
+
+	return res;
+}
+
 class CSearchDialogFileList : public CFileListCtrl<CSearchFileData>
 {
 	friend class CSearchDialog;
@@ -44,9 +72,9 @@ protected:
 #endif
 
 private:
-	virtual bool CanStartComparison(wxString*) { return false; }
+	virtual bool CanStartComparison() { return false; }
 	virtual void StartComparison() {}
-	virtual bool GetNextFile(wxString&, bool &, wxLongLong &, CDateTime&) { return false; }
+	virtual bool get_next_file(wxString&, bool &, int64_t&, fz::datetime&) { return false; }
 	virtual void CompareAddFile(CComparableListing::t_fileEntryFlags) {}
 	virtual void FinishComparison() {}
 	virtual void ScrollTopItem(int) {}
@@ -94,7 +122,7 @@ bool CSearchDialogFileList::ItemIsDir(int index) const
 
 int64_t CSearchDialogFileList::ItemGetSize(int index) const
 {
-	return m_fileData[index].size.GetValue();
+	return m_fileData[index].size;
 }
 
 CFileListCtrl<CSearchFileData>::CSortComparisonObject CSearchDialogFileList::GetSortComparisonObject()
@@ -116,7 +144,7 @@ CFileListCtrl<CSearchFileData>::CSortComparisonObject CSearchDialogFileList::Get
 		else if (m_sortColumn == 6)
 			return CFileListCtrl<CSearchFileData>::CSortComparisonObject(new CFileListCtrlSortOwnerGroup<std::vector<CSearchFileData>, CSearchFileData>(m_fileData, m_fileData, dirSortMode, nameSortMode, this));
 		else
-			return CFileListCtrl<CSearchFileData>::CSortComparisonObject(new CFileListCtrlSortName<std::vector<CSearchFileData>, CSearchFileData>(m_fileData, m_fileData, dirSortMode, nameSortMode, this));
+			return CFileListCtrl<CSearchFileData>::CSortComparisonObject(new CFileListCtrlSortNamePath<std::vector<CSearchFileData>, CSearchFileData>(m_fileData, m_fileData, dirSortMode, nameSortMode, this));
 	}
 	else {
 		if (m_sortColumn == 1)
@@ -132,7 +160,7 @@ CFileListCtrl<CSearchFileData>::CSortComparisonObject CSearchDialogFileList::Get
 		else if (m_sortColumn == 6)
 			return CFileListCtrl<CSearchFileData>::CSortComparisonObject(new CReverseSort<CFileListCtrlSortOwnerGroup<std::vector<CSearchFileData>, CSearchFileData>, CSearchFileData>(m_fileData, m_fileData, dirSortMode, nameSortMode, this));
 		else
-			return CFileListCtrl<CSearchFileData>::CSortComparisonObject(new CReverseSort<CFileListCtrlSortName<std::vector<CSearchFileData>, CSearchFileData>, CSearchFileData>(m_fileData, m_fileData, dirSortMode, nameSortMode, this));
+			return CFileListCtrl<CSearchFileData>::CSortComparisonObject(new CReverseSort<CFileListCtrlSortNamePath<std::vector<CSearchFileData>, CSearchFileData>, CSearchFileData>(m_fileData, m_fileData, dirSortMode, nameSortMode, this));
 	}
 }
 
@@ -151,7 +179,7 @@ wxString CSearchDialogFileList::GetItemText(int item, unsigned int column)
 		if (entry.is_dir() || entry.size < 0)
 			return wxString();
 		else
-			return CSizeFormat::Format(entry.size.GetValue());
+			return CSizeFormat::Format(entry.size);
 	}
 	else if (column == 3) {
 		CSearchFileData& data = m_fileData[index];
@@ -225,6 +253,7 @@ bool CSearchDialog::Load()
 	/* XRCed complains if adding a status bar to a dialog, so do it here instead */
 	CFilelistStatusBar* pStatusBar = new CFilelistStatusBar(this);
 	pStatusBar->SetEmptyString(_("No search results"));
+	pStatusBar->SetConnected(true);
 
 	GetSizer()->Add(pStatusBar, 0, wxGROW);
 
@@ -262,19 +291,15 @@ void CSearchDialog::Run()
 	m_original_dir = m_pState->GetRemotePath();
 	m_local_target = m_pState->GetLocalDir();
 
-	m_pState->BlockHandlers(STATECHANGE_REMOTE_DIR);
-	m_pState->BlockHandlers(STATECHANGE_REMOTE_DIR_MODIFIED);
-	m_pState->RegisterHandler(this, STATECHANGE_REMOTE_DIR, false);
-	m_pState->RegisterHandler(this, STATECHANGE_REMOTE_IDLE, false);
+	m_pState->RegisterHandler(this, STATECHANGE_REMOTE_DIR_OTHER);
+	m_pState->RegisterHandler(this, STATECHANGE_REMOTE_IDLE);
 
 	ShowModal();
 
 	SaveConditions();
 
 	m_pState->UnregisterHandler(this, STATECHANGE_REMOTE_IDLE);
-	m_pState->UnregisterHandler(this, STATECHANGE_REMOTE_DIR);
-	m_pState->UnblockHandlers(STATECHANGE_REMOTE_DIR);
-	m_pState->UnblockHandlers(STATECHANGE_REMOTE_DIR_MODIFIED);
+	m_pState->UnregisterHandler(this, STATECHANGE_REMOTE_DIR_OTHER);
 
 	if (m_searching) {
 		if (!m_pState->IsRemoteIdle()) {
@@ -290,10 +315,15 @@ void CSearchDialog::Run()
 	}
 }
 
-void CSearchDialog::OnStateChange(CState* pState, enum t_statechange_notifications notification, const wxString& data, const void* data2)
+void CSearchDialog::OnStateChange(CState* pState, enum t_statechange_notifications notification, const wxString&, const void* data2)
 {
-	if (notification == STATECHANGE_REMOTE_DIR)
-		ProcessDirectoryListing();
+	if (notification == STATECHANGE_REMOTE_DIR_OTHER && data2) {
+		auto recursiveOperation = m_pState->GetRecursiveOperationHandler();
+		if (recursiveOperation && recursiveOperation->GetOperationMode() == CRecursiveOperation::recursive_list) {
+			std::shared_ptr<CDirectoryListing> const& listing = *reinterpret_cast<std::shared_ptr<CDirectoryListing> const*>(data2);
+			ProcessDirectoryListing(listing);
+		}
+	}
 	else if (notification == STATECHANGE_REMOTE_IDLE) {
 		if (pState->IsRemoteIdle())
 			m_searching = false;
@@ -301,10 +331,8 @@ void CSearchDialog::OnStateChange(CState* pState, enum t_statechange_notificatio
 	}
 }
 
-void CSearchDialog::ProcessDirectoryListing()
+void CSearchDialog::ProcessDirectoryListing(std::shared_ptr<CDirectoryListing> const& listing)
 {
-	std::shared_ptr<CDirectoryListing> listing = m_pState->GetRemoteDir();
-
 	if (!listing || listing->failed())
 		return;
 
@@ -334,7 +362,7 @@ void CSearchDialog::ProcessDirectoryListing()
 		if (entry.is_dir())
 			m_results->GetFilelistStatusBar()->AddDirectory();
 		else
-			m_results->GetFilelistStatusBar()->AddFile(entry.size.GetValue());
+			m_results->GetFilelistStatusBar()->AddFile(entry.size);
 	}
 
 	if (added) {
@@ -344,7 +372,7 @@ void CSearchDialog::ProcessDirectoryListing()
 	}
 }
 
-void CSearchDialog::OnSearch(wxCommandEvent& event)
+void CSearchDialog::OnSearch(wxCommandEvent&)
 {
 	if (!m_pState->IsRemoteIdle()) {
 		wxBell();
@@ -393,12 +421,14 @@ void CSearchDialog::OnSearch(wxCommandEvent& event)
 
 	// Start
 	m_searching = true;
-	m_pState->GetRecursiveOperationHandler()->AddDirectoryToVisitRestricted(path, _T(""), true);
+	recursion_root root(path, true);
+	root.add_dir_to_visit_restricted(path, _T(""), true);
+	m_pState->GetRecursiveOperationHandler()->AddRecursionRoot(std::move(root));
 	std::vector<CFilter> const filters; // Empty, recurse into everything
-	m_pState->GetRecursiveOperationHandler()->StartRecursiveOperation(CRecursiveOperation::recursive_list, path, filters, true);
+	m_pState->GetRecursiveOperationHandler()->StartRecursiveOperation(CRecursiveOperation::recursive_list, filters, path);
 }
 
-void CSearchDialog::OnStop(wxCommandEvent& event)
+void CSearchDialog::OnStop(wxCommandEvent&)
 {
 	if (!m_pState->IsRemoteIdle()) {
 		m_pState->m_pCommandQueue->Cancel();
@@ -427,6 +457,7 @@ void CSearchDialog::OnContextMenu(wxContextMenuEvent& event)
 	if (!m_pState->IsRemoteIdle()) {
 		pMenu->Enable(XRCID("ID_MENU_SEARCH_DOWNLOAD"), false);
 		pMenu->Enable(XRCID("ID_MENU_SEARCH_DELETE"), false);
+		pMenu->Enable(XRCID("ID_MENU_SEARCH_EDIT"), false);
 	}
 
 	PopupMenu(pMenu);
@@ -475,7 +506,7 @@ EVT_BUTTON(XRCID("ID_BROWSE"), CSearchDownloadDialog::OnBrowse)
 EVT_BUTTON(XRCID("wxID_OK"), CSearchDownloadDialog::OnOK)
 END_EVENT_TABLE()
 
-void CSearchDownloadDialog::OnBrowse(wxCommandEvent& event)
+void CSearchDownloadDialog::OnBrowse(wxCommandEvent&)
 {
 	wxTextCtrl *pText = XRCCTRL(*this, "ID_LOCALPATH", wxTextCtrl);
 
@@ -484,7 +515,7 @@ void CSearchDownloadDialog::OnBrowse(wxCommandEvent& event)
 		pText->ChangeValue(dlg.GetPath());
 }
 
-void CSearchDownloadDialog::OnOK(wxCommandEvent& event)
+void CSearchDownloadDialog::OnOK(wxCommandEvent&)
 {
 	wxTextCtrl *pText = XRCCTRL(*this, "ID_LOCALPATH", wxTextCtrl);
 
@@ -502,8 +533,10 @@ void CSearchDownloadDialog::OnOK(wxCommandEvent& event)
 	EndDialog(wxID_OK);
 }
 
-void CSearchDialog::ProcessSelection(std::list<int> &selected_files, std::list<CServerPath> &selected_dirs)
+void CSearchDialog::ProcessSelection(std::list<int> &selected_files, std::deque<CServerPath> &selected_dirs)
 {
+	std::deque<CServerPath> dirs;
+
 	int sel = -1;
 	while ((sel = m_results->GetNextItem(sel, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED)) != -1) {
 		if (sel > (int)m_results->m_indexMapping.size())
@@ -516,55 +549,34 @@ void CSearchDialog::ProcessSelection(std::list<int> &selected_files, std::list<C
 			if (path.empty())
 				continue;
 
-			bool replaced = false;
-			std::list<CServerPath>::iterator iter = selected_dirs.begin();
-			std::list<CServerPath>::iterator prev;
-
-			// Make sure that selected_dirs does not contain
-			// any directories that are in a parent-child relationship
-			// Resolve by only keeping topmost parents
-			while (iter != selected_dirs.end()) {
-				if (*iter == path) {
-					replaced = true;
-					break;
-				}
-
-				if (iter->IsParentOf(path, false)) {
-					replaced = true;
-					break;
-				}
-
-				if (iter->IsSubdirOf(path, false)) {
-					if (!replaced) {
-						*iter = path;
-						replaced = true;
-					}
-					else {
-						prev = iter++;
-						selected_dirs.erase(prev);
-						continue;
-					}
-				}
-				++iter;
-			}
-			if (!replaced)
-				selected_dirs.push_back(path);
+			dirs.push_back(path);
 		}
 		else
 			selected_files.push_back(index);
 	}
 
+	// Make sure that selected_dirs does not contain
+	// any directories that are in a parent-child relationship
+	// Resolve by only keeping topmost parents
+	std::sort(dirs.begin(), dirs.end());
+	for (CServerPath const& path : dirs) {
+		if (!selected_dirs.empty() && (path.IsSubdirOf(selected_dirs.back(), false) || path == selected_dirs.back())) {
+			continue;
+		}
+		selected_dirs.push_back(path);
+	}
+
 	// Now in a second phase filter out all files that are also in a directory
 	std::list<int> selected_files_new;
-	for (auto const& sel : selected_files) {
-		CServerPath path = m_results->m_fileData[sel].path;
-		std::list<CServerPath>::const_iterator path_iter;
+	for (auto const& sel_file : selected_files) {
+		CServerPath const& path = m_results->m_fileData[sel_file].path;
+		std::deque<CServerPath>::iterator path_iter;
 		for (path_iter = selected_dirs.begin(); path_iter != selected_dirs.end(); ++path_iter) {
 			if (*path_iter == path || path_iter->IsParentOf(path, false))
 				break;
 		}
 		if (path_iter == selected_dirs.end())
-			selected_files_new.push_back(sel);
+			selected_files_new.push_back(sel_file);
 	}
 	selected_files.swap(selected_files_new);
 
@@ -579,18 +591,13 @@ void CSearchDialog::OnDownload(wxCommandEvent&)
 		return;
 
 	// Find all selected files and directories
-	std::list<CServerPath> selected_dirs;
+	std::deque<CServerPath> selected_dirs;
 	std::list<int> selected_files;
 	ProcessSelection(selected_files, selected_dirs);
 
 	if (selected_files.empty() && selected_dirs.empty())
 		return;
 
-	if (selected_dirs.size() > 1) {
-		wxMessageBoxEx(_("Downloading multiple unrelated directories is not yet supported"), _("Downloading search results"), wxICON_EXCLAMATION);
-		return;
-	}
-
 	CSearchDownloadDialog dlg;
 	if (!dlg.Run(this, m_local_target.GetPath(), selected_files.size(), selected_dirs.size()))
 		return;
@@ -637,7 +644,7 @@ void CSearchDialog::OnDownload(wxCommandEvent&)
 
 		m_pQueue->QueueFile(!start, true,
 			entry.name, (localName != entry.name) ? localName : wxString(),
-			target_path, remote_path, *pServer, entry.size.GetValue());
+			target_path, remote_path, *pServer, entry.size);
 	}
 	m_pQueue->QueueFile_Finish(start);
 
@@ -652,10 +659,12 @@ void CSearchDialog::OnDownload(wxCommandEvent&)
 		if (!flatten && dir.HasParent())
 			target_path.AddSegment(dir.GetLastSegment());
 
-		m_pState->GetRecursiveOperationHandler()->AddDirectoryToVisit(dir, _T(""), target_path, false);
-		std::vector<CFilter> const filters; // Empty, recurse into everything
-		m_pState->GetRecursiveOperationHandler()->StartRecursiveOperation(mode, dir, filters, true, m_original_dir);
+		recursion_root root(dir, true);
+		root.add_dir_to_visit(dir, _T(""), target_path, false);
+		m_pState->GetRecursiveOperationHandler()->AddRecursionRoot(std::move(root));
 	}
+	std::vector<CFilter> const filters; // Empty, recurse into everything
+	m_pState->GetRecursiveOperationHandler()->StartRecursiveOperation(mode, filters, m_original_dir);
 }
 
 void CSearchDialog::OnEdit(wxCommandEvent&)
@@ -664,7 +673,7 @@ void CSearchDialog::OnEdit(wxCommandEvent&)
 		return;
 
 	// Find all selected files and directories
-	std::list<CServerPath> selected_dirs;
+	std::deque<CServerPath> selected_dirs;
 	std::list<int> selected_files;
 	ProcessSelection(selected_files, selected_dirs);
 
@@ -707,7 +716,7 @@ void CSearchDialog::OnEdit(wxCommandEvent&)
 		const CDirentry& entry = m_results->m_fileData[item];
 		const CServerPath path = m_results->m_fileData[item].path;
 
-		pEditHandler->Edit(CEditHandler::remote, entry.name, path, *pServer, entry.size.GetValue(), this);
+		pEditHandler->Edit(CEditHandler::remote, entry.name, path, *pServer, entry.size, this);
 	}
 }
 
@@ -717,18 +726,13 @@ void CSearchDialog::OnDelete(wxCommandEvent&)
 		return;
 
 	// Find all selected files and directories
-	std::list<CServerPath> selected_dirs;
+	std::deque<CServerPath> selected_dirs;
 	std::list<int> selected_files;
 	ProcessSelection(selected_files, selected_dirs);
 
 	if (selected_files.empty() && selected_dirs.empty())
 		return;
 
-	if (selected_dirs.size() > 1) {
-		wxMessageBoxEx(_("Deleting multiple unrelated directories is not yet supported"), _("Deleting directories"), wxICON_EXCLAMATION);
-		return;
-	}
-
 	wxString question;
 	if (selected_dirs.empty())
 		question.Printf(wxPLURAL("Really delete %d file from the server?", "Really delete %d files from the server?", selected_files.size()), selected_files.size());
@@ -751,16 +755,17 @@ void CSearchDialog::OnDelete(wxCommandEvent&)
 	}
 
 	for (auto path : selected_dirs) {
-		if (!path.HasParent())
-			m_pState->GetRecursiveOperationHandler()->AddDirectoryToVisit(path, _T(""));
-		else {
-			m_pState->GetRecursiveOperationHandler()->AddDirectoryToVisit(path.GetParent(), path.GetLastSegment());
+		wxString segment;
+		if (path.HasParent()) {
+			segment = path.GetLastSegment();
 			path = path.GetParent();
 		}
-
-		std::vector<CFilter> const filters; // Empty, recurse into everything
-		m_pState->GetRecursiveOperationHandler()->StartRecursiveOperation(CRecursiveOperation::recursive_delete, path, filters, !path.HasParent(), m_original_dir);
+		recursion_root root(path, !path.HasParent());
+		root.add_dir_to_visit(path, segment);
+		m_pState->GetRecursiveOperationHandler()->AddRecursionRoot(std::move(root));
 	}
+	std::vector<CFilter> const filters; // Empty, recurse into everything
+	m_pState->GetRecursiveOperationHandler()->StartRecursiveOperation(CRecursiveOperation::recursive_delete, filters, m_original_dir);
 }
 
 void CSearchDialog::OnCharHook(wxKeyEvent& event)
@@ -792,17 +797,17 @@ void CSearchDialog::LoadConditions()
 	CInterProcessMutex mutex(MUTEX_SEARCHCONDITIONS);
 
 	CXmlFile file(wxGetApp().GetSettingsFile(_T("search")));
-	TiXmlElement* pDocument = file.Load();
-	if (!pDocument) {
+	auto document = file.Load();
+	if (!document) {
 		wxMessageBoxEx(file.GetError(), _("Error loading xml file"), wxICON_ERROR);
 		return;
 	}
 
-	TiXmlElement* pFilter = pDocument->FirstChildElement("Filter");
-	if (!pFilter)
+	auto filter = document.child("Filter");
+	if (!filter)
 		return;
 
-	if (!CFilterManager::LoadFilter(pFilter, m_search_filter))
+	if (!CFilterManager::LoadFilter(filter, m_search_filter))
 		m_search_filter = CFilter();
 }
 
@@ -811,18 +816,18 @@ void CSearchDialog::SaveConditions()
 	CInterProcessMutex mutex(MUTEX_SEARCHCONDITIONS);
 
 	CXmlFile file(wxGetApp().GetSettingsFile(_T("search")));
-	TiXmlElement* pDocument = file.Load();
-	if (!pDocument) {
+	auto document = file.Load();
+	if (!document) {
 		wxMessageBoxEx(file.GetError(), _("Error loading xml file"), wxICON_ERROR);
 		return;
 	}
 
-	TiXmlElement* pFilter;
-	while ((pFilter = pDocument->FirstChildElement("Filter")))
-		pDocument->RemoveChild(pFilter);
-	pFilter = pDocument->LinkEndChild(new TiXmlElement("Filter"))->ToElement();
+	pugi::xml_node filter;
+	while ((filter = document.child("Filter")))
+		document.remove_child(filter);
+	filter = document.append_child("Filter");
 
-	CFilterDialog::SaveFilter(pFilter, m_search_filter);
+	CFilterDialog::SaveFilter(filter, m_search_filter);
 
 	file.Save(true);
 }
