@@ -7,8 +7,10 @@
 #include "queue.h"
 #include "filezillaapp.h"
 #include "recursive_operation.h"
-#include "local_filesys.h"
 #include "listingcomparison.h"
+#include "xrc_helper.h"
+
+#include <libfilezilla/local_filesys.hpp>
 
 CContextManager CContextManager::m_the_context_manager;
 
@@ -270,6 +272,7 @@ bool CState::SetLocalDir(CLocalPath const& dir, wxString *error, bool rememberPr
 
 			m_sync_browse.is_changing = true;
 			m_sync_browse.compare = m_pComparisonManager->IsComparing();
+			m_sync_browse.target_path = remote_path;
 			CListCommand *pCommand = new CListCommand(remote_path);
 			m_pCommandQueue->ProcessCommand(pCommand);
 
@@ -645,20 +648,17 @@ void CState::UploadDroppedFiles(const wxFileDataObject* pFileDataObject, const C
 
 	const wxArrayString& files = pFileDataObject->GetFilenames();
 
-	for (unsigned int i = 0; i < files.Count(); i++)
-	{
+	for (unsigned int i = 0; i < files.Count(); ++i) {
 		int64_t size;
 		bool is_link;
-		CLocalFileSystem::local_fileType type = CLocalFileSystem::GetFileInfo(files[i], is_link, &size, 0, 0);
-		if (type == CLocalFileSystem::file)
-		{
+		fz::local_filesys::type type = fz::local_filesys::get_file_info(fz::to_native(files[i]), is_link, &size, 0, 0);
+		if (type == fz::local_filesys::file) {
 			wxString localFile;
 			const CLocalPath localPath(files[i], &localFile);
 			m_mainFrame.GetQueue()->QueueFile(queueOnly, false, localFile, wxEmptyString, localPath, path, *m_pServer, size);
 			m_mainFrame.GetQueue()->QueueFile_Finish(!queueOnly);
 		}
-		else if (type == CLocalFileSystem::dir)
-		{
+		else if (type == fz::local_filesys::dir) {
 			CLocalPath localPath(files[i]);
 			if (localPath.HasParent())
 			{
@@ -679,15 +679,15 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, const C
 #ifdef __WXMSW__
 	int len = 1;
 
-	for (unsigned int i = 0; i < files.Count(); i++)
+	for (unsigned int i = 0; i < files.Count(); ++i) {
 		len += files[i].Len() + 1;
+	}
 
 	// SHFILEOPSTRUCT's pTo and pFrom accept null-terminated lists
 	// of null-terminated filenames.
 	wxChar* from = new wxChar[len];
 	wxChar* p = from;
-	for (unsigned int i = 0; i < files.Count(); i++)
-	{
+	for (unsigned int i = 0; i < files.Count(); ++i) {
 		wxStrcpy(p, files[i]);
 		p += files[i].Len() + 1;
 	}
@@ -708,15 +708,13 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, const C
 	delete [] from;
 #else
 	wxString error;
-	for (unsigned int i = 0; i < files.Count(); i++)
-	{
-		const wxString& file(files[i]);
+	for (unsigned int i = 0; i < files.Count(); ++i) {
+		wxString const& file(files[i]);
 
 		int64_t size;
 		bool is_link;
-		CLocalFileSystem::local_fileType type = CLocalFileSystem::GetFileInfo(file, is_link, &size, 0, 0);
-		if (type == CLocalFileSystem::file)
-		{
+		fz::local_filesys::type type = fz::local_filesys::get_file_info(fz::to_native(file), is_link, &size, 0, 0);
+		if (type == fz::local_filesys::file) {
 			wxString name;
 			CLocalPath sourcePath(file, &name);
 			if (name.empty())
@@ -730,21 +728,18 @@ void CState::HandleDroppedFiles(const wxFileDataObject* pFileDataObject, const C
 			else
 				wxRenameFile(file, target);
 		}
-		else if (type == CLocalFileSystem::dir)
-		{
+		else if (type == fz::local_filesys::dir) {
 			CLocalPath sourcePath(file);
 			if (sourcePath == path || sourcePath.GetParent() == path)
 				continue;
-			if (sourcePath.IsParentOf(path))
-			{
+			if (sourcePath.IsParentOf(path)) {
 				error = _("A directory cannot be dragged into one of its subdirectories.");
 				continue;
 			}
 
 			if (copy)
 				RecursiveCopy(sourcePath, path);
-			else
-			{
+			else {
 				if (!sourcePath.HasParent())
 					continue;
 				wxRenameFile(file, path.GetPath() + sourcePath.GetLastSegment());
@@ -780,28 +775,24 @@ bool CState::RecursiveCopy(CLocalPath source, const CLocalPath& target)
 	dirsToVisit.push_back(last_segment + CLocalPath::path_separator);
 
 	// Process any subdirs which still have to be visited
-	while (!dirsToVisit.empty())
-	{
+	while (!dirsToVisit.empty()) {
 		wxString dirname = dirsToVisit.front();
 		dirsToVisit.pop_front();
 		wxMkdir(target.GetPath() + dirname);
 
-		CLocalFileSystem fs;
-		if (!fs.BeginFindFiles(source.GetPath() + dirname, false))
+		fz::local_filesys fs;
+		if (!fs.begin_find_files(fz::to_native(source.GetPath() + dirname), false))
 			continue;
 
 		bool is_dir, is_link;
-		wxString file;
-		while (fs.GetNextFile(file, is_link, is_dir, 0, 0, 0))
-		{
-			if (file.empty())
-			{
+		fz::native_string file;
+		while (fs.get_next_file(file, is_link, is_dir, 0, 0, 0)) {
+			if (file.empty()) {
 				wxGetApp().DisplayEncodingWarning();
 				continue;
 			}
 
-			if (is_dir)
-			{
+			if (is_dir) {
 				if (is_link)
 					continue;
 
@@ -821,8 +812,7 @@ bool CState::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, co
 	bool hasDirs = false;
 	bool hasFiles = false;
 	const std::list<CRemoteDataObject::t_fileInfo>& files = pRemoteDataObject->GetFiles();
-	for (std::list<CRemoteDataObject::t_fileInfo>::const_iterator iter = files.begin(); iter != files.end(); ++iter)
-	{
+	for (std::list<CRemoteDataObject::t_fileInfo>::const_iterator iter = files.begin(); iter != files.end(); ++iter) {
 		if (iter->dir)
 			hasDirs = true;
 		else
@@ -840,20 +830,26 @@ bool CState::DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, co
 	if (!hasDirs)
 		return true;
 
+	recursion_root root(pRemoteDataObject->GetServerPath(), false);
 	for (std::list<CRemoteDataObject::t_fileInfo>::const_iterator iter = files.begin(); iter != files.end(); ++iter) {
 		if (!iter->dir)
 			continue;
 
 		CLocalPath newPath(path);
 		newPath.AddSegment(CQueueView::ReplaceInvalidCharacters(iter->name));
-		m_pRecursiveOperation->AddDirectoryToVisit(pRemoteDataObject->GetServerPath(), iter->name, newPath, iter->link);
+		root.add_dir_to_visit(pRemoteDataObject->GetServerPath(), iter->name, newPath, iter->link);
 	}
 
-	if (m_pComparisonManager->IsComparing())
-		m_pComparisonManager->ExitComparisonMode();
+	if (!root.empty()) {
+		if (m_pComparisonManager->IsComparing())
+			m_pComparisonManager->ExitComparisonMode();
 
-	CFilterManager filter;
-	m_pRecursiveOperation->StartRecursiveOperation(queueOnly ? CRecursiveOperation::recursive_addtoqueue : CRecursiveOperation::recursive_download, pRemoteDataObject->GetServerPath(), filter.GetActiveFilters(false));
+		m_pRecursiveOperation->AddRecursionRoot(std::move(root));
+
+		CFilterManager filter;
+		m_pRecursiveOperation->StartRecursiveOperation(queueOnly ? CRecursiveOperation::recursive_addtoqueue : CRecursiveOperation::recursive_download,
+			filter.GetActiveFilters(false), pRemoteDataObject->GetServerPath());
+	}
 
 	return true;
 }
@@ -877,8 +873,35 @@ bool CState::IsRemoteIdle(bool ignore_recursive) const
 	return m_pCommandQueue->Idle(ignore_recursive ? CCommandQueue::normal : CCommandQueue::any);
 }
 
-void CState::ListingFailed(int error)
+void CState::ListingFailed(int)
 {
+	if (m_sync_browse.is_changing && !m_sync_browse.target_path.empty()) {
+		wxDialogEx dlg;
+		if (!dlg.Load(0, _T("ID_SYNCBROWSE_NONEXISTING"))) {
+			SetSyncBrowse(false);
+			return;
+		}
+
+		xrc_call(dlg, "ID_SYNCBROWSE_NONEXISTING_LABEL", &wxStaticText::SetLabel, wxString::Format(_("The remote directory '%s' does not exist."), m_sync_browse.target_path.GetPath()));
+		xrc_call(dlg, "ID_SYNCBROWSE_CREATE", &wxRadioButton::SetLabel, _("Create &missing remote directory and enter it"));
+		xrc_call(dlg, "ID_SYNCBROWSE_DISABLE", &wxRadioButton::SetLabel, _("&Disable synchronized browsing and continue changing the local directory"));
+		dlg.GetSizer()->Fit(&dlg);
+		if (dlg.ShowModal() == wxID_OK) {
+
+			if (xrc_call(dlg, "ID_SYNCBROWSE_CREATE", &wxRadioButton::GetValue)) {
+				m_pCommandQueue->ProcessCommand(new CMkdirCommand(m_sync_browse.target_path));
+				m_pCommandQueue->ProcessCommand(new CListCommand(m_sync_browse.target_path));
+				m_sync_browse.target_path.clear();
+				return;
+			}
+			else {
+				CLocalPath local = GetSynchronizedDirectory(m_sync_browse.target_path);
+				SetSyncBrowse(false);
+				SetLocalDir(local);
+			}
+		}
+	}
+
 	m_sync_browse.is_changing = false;
 }
 
@@ -927,13 +950,41 @@ bool CState::ChangeRemoteDir(const CServerPath& path, const wxString& subdir /*=
 			}
 			else if (!local_path.Exists(&error)) {
 				wxString msg = error + _T("\n") + _("Disable synchronized browsing and continue changing the remote directory?");
-				if (wxMessageBoxEx(msg, _("Synchronized browsing"), wxICON_QUESTION | wxYES_NO) != wxYES)
+
+				wxDialogEx dlg;
+				if (!dlg.Load(0, _T("ID_SYNCBROWSE_NONEXISTING"))) {
 					return false;
-				SetSyncBrowse(false);
+				}
+				xrc_call(dlg, "ID_SYNCBROWSE_NONEXISTING_LABEL", &wxStaticText::SetLabel, wxString::Format(_("The local directory '%s' does not exist."), local_path.GetPath()));
+				xrc_call(dlg, "ID_SYNCBROWSE_CREATE", &wxRadioButton::SetLabel, _("Create &missing local directory and enter it"));
+				xrc_call(dlg, "ID_SYNCBROWSE_DISABLE", &wxRadioButton::SetLabel, _("&Disable synchronized browsing and continue changing the remote directory"));
+				dlg.GetSizer()->Fit(&dlg);
+				if (dlg.ShowModal() != wxID_OK) {
+					return false;
+				}
+
+				if (xrc_call(dlg, "ID_SYNCBROWSE_CREATE", &wxRadioButton::GetValue)) {
+					{
+						wxLogNull log;
+						wxMkdir(local_path.GetPath());
+					}
+
+					if (!local_path.Exists(&error)) {
+						wxMessageBox(wxString::Format(_("The local directory '%s' could not be created."), local_path.GetPath()), _("Synchronized browsing"), wxICON_EXCLAMATION);
+						return false;
+					}
+					m_sync_browse.is_changing = true;
+					m_sync_browse.compare = m_pComparisonManager->IsComparing();
+					m_sync_browse.target_path.clear();
+				}
+				else {
+					SetSyncBrowse(false);
+				}
 			}
 			else {
 				m_sync_browse.is_changing = true;
 				m_sync_browse.compare = m_pComparisonManager->IsComparing();
+				m_sync_browse.target_path.clear();
 			}
 		}
 	}
@@ -949,8 +1000,7 @@ bool CState::SetSyncBrowse(bool enable, const CServerPath& assumed_remote_root /
 	if (enable != m_sync_browse.local_root.empty())
 		return enable;
 
-	if (!enable)
-	{
+	if (!enable) {
 		wxASSERT(assumed_remote_root.empty());
 		m_sync_browse.local_root.clear();
 		m_sync_browse.remote_root.clear();
@@ -960,8 +1010,7 @@ bool CState::SetSyncBrowse(bool enable, const CServerPath& assumed_remote_root /
 		return false;
 	}
 
-	if (!m_pDirectoryListing && assumed_remote_root.empty())
-	{
+	if (!m_pDirectoryListing && assumed_remote_root.empty()) {
 		NotifyHandlers(STATECHANGE_SYNC_BROWSE);
 		return false;
 	}
@@ -971,8 +1020,7 @@ bool CState::SetSyncBrowse(bool enable, const CServerPath& assumed_remote_root /
 
 	if (assumed_remote_root.empty())
 		m_sync_browse.remote_root = m_pDirectoryListing->path;
-	else
-	{
+	else {
 		m_sync_browse.remote_root = assumed_remote_root;
 		m_sync_browse.is_changing = true;
 		m_sync_browse.compare = false;
@@ -992,8 +1040,7 @@ bool CState::SetSyncBrowse(bool enable, const CServerPath& assumed_remote_root /
 CLocalPath CState::GetSynchronizedDirectory(CServerPath remote_path)
 {
 	std::list<wxString> segments;
-	while (remote_path.HasParent() && remote_path != m_sync_browse.remote_root)
-	{
+	while (remote_path.HasParent() && remote_path != m_sync_browse.remote_root) {
 		segments.push_front(remote_path.GetLastSegment());
 		remote_path = remote_path.GetParent();
 	}
@@ -1011,8 +1058,7 @@ CLocalPath CState::GetSynchronizedDirectory(CServerPath remote_path)
 CServerPath CState::GetSynchronizedDirectory(CLocalPath local_path)
 {
 	std::list<wxString> segments;
-	while (local_path.HasParent() && local_path != m_sync_browse.local_root)
-	{
+	while (local_path.HasParent() && local_path != m_sync_browse.local_root) {
 		wxString last;
 		local_path.MakeParent(&last);
 		segments.push_front(last);
